@@ -94,17 +94,34 @@ sub cmd_inject {
     $self->install("--reinstall", @args);
 }
 
+sub cmd_update {
+    my($self, @args) = @_;
+
+    die "Usage: carmel update\n" if @args; # TODO supprot args
+
+    # reinstall everything in cpanfile to build artifacts
+    my $cpanfile = $self->try_cpanfile
+      or die "Can't locate 'cpanfile' to load module list.\n";
+    $self->install_with_cpanfile(Module::CPANfile->load($cpanfile));
+
+    # then rebuild the snapshot
+    my @artifacts = $self->install_from_cpanfile(1);
+    $self->dump_bootstrap(\@artifacts);
+    $self->save_snapshot(\@artifacts);
+}
+
 sub cmd_install {
     my($self, @args) = @_;
 
     die "Usage: carmel install\n" if @args;
 
-    $self->install_from_cpanfile;
-    $self->post_install;
+    my @artifacts = $self->install_from_cpanfile;
+    $self->dump_bootstrap(\@artifacts);
+    $self->save_snapshot(\@artifacts);
 }
 
 sub install_from_cpanfile {
-    my $self = shift;
+    my($self, $no_snapshot) = @_;
 
     my $requirements = CPAN::Meta::Requirements->new;
     $self->resolve(
@@ -113,9 +130,16 @@ sub install_from_cpanfile {
             printf "Using %s (%s)\n", $artifact->package, $artifact->version || '0';
         },
         sub {
-            my($module, $want_version) = @_;
+            my($module, $want_version, $dist) = @_;
+            if ($dist) {
+                # TODO pass $dist->distfile to cpanfile
+                my $ver = version::->parse($dist->version_for($module));
+                $want_version = $ver > 0 ? "== $ver" : $ver;
+            }
             $requirements->add_string_requirement($module => $want_version);
         },
+        1, # strict
+        $no_snapshot,
     );
 
     if (my @missing = $requirements->required_modules) {
@@ -125,33 +149,44 @@ sub install_from_cpanfile {
             },
         });
         print "---> Installing new dependencies: ", join(", ", @missing), "\n";
-        $self->install_with_cpanfile($cpanfile);
+        $self->install_with_cpanfile($cpanfile, $no_snapshot ? () : $self->snapshot );
     }
 
     my @artifacts;
-    $self->resolve(sub { push @artifacts, $_[0] });
+    $self->resolve(sub { push @artifacts, $_[0] }, undef, 0, $no_snapshot);
 
     # $self->requirements has been upgraded at this point with the whole subreqs
     printf "---> Complete! %d cpanfile dependencies. %d modules installed.\n" .
       "---> Use `carmel show [module]` to see where a module is installed.\n",
       scalar(grep { $_ ne 'perl' } $self->build_requirements->required_modules), scalar(@artifacts);
+
+    return @artifacts;
 }
 
 sub is_core {
     my($self, $module, $want_version) = @_;
-
     return unless exists $Module::CoreList::version{$]+0}{$module};
-
-    CPAN::Meta::Requirements->from_string_hash({ $module => $want_version })
-        ->accepts_module($module, $Module::CoreList::version{$]+0}{$module} || '0');
+    $self->accepts($module, $want_version, $Module::CoreList::version{$]+0}{$module});
 }
 
 sub install_with_cpanfile {
-    my($self, $cpanfile) = @_;
+    my($self, $cpanfile, $snapshot) = @_;
 
     my $path = Path::Tiny->tempfile;
     $cpanfile->save($path);
-    $self->install("--installdeps", "--cpanfile", $path, ".");
+
+    my @options = ("--cpanfile", $path);
+
+    if ($snapshot) {
+        my $path = Path::Tiny->tempfile;
+        $snapshot->write_index($path);
+        push @options,
+          "--mirror-index", $path,
+          "--cascade-search",
+          "--mirror", "http://backpan.perl.org";
+    }
+
+    $self->install("--installdeps", @options, ".");
 }
 
 sub install {
@@ -165,10 +200,17 @@ sub install {
     local $ENV{PERL_CPANM_HOME} = $dir;
     local $ENV{PERL_CPANM_OPT};
 
+    my $cpanfile = $self->try_cpanfile
+      or die "Can't locate 'cpanfile' to load module list.\n";
+
+    # one mirror for now
+    my $mirror = Module::CPANfile->load($cpanfile)->mirrors->[0];
+
     require Menlo::CLI::Compat;
 
     my $cli = Menlo::CLI::Compat->new(
         ($self->verbose ? () : "--quiet"),
+        ($mirror ? ("-M", $mirror) : ()),
         "--notest",
         "--save-dists", $self->repository_base->child('cache'),
         "-L", $self->repository_base->child('perl5'),
@@ -180,6 +222,8 @@ sub install {
         next unless $ent->is_dir && $ent->child("blib/meta/install.json")->exists;
         $self->repo->import_artifact($ent);
     }
+
+    $self->repository_base->child('perl5')->remove_tree({ safe => 0 });
 }
 
 sub quote {
@@ -210,16 +254,6 @@ sub transform {
     }
 }
 
-sub post_install {
-    my $self = shift;
-
-    my @artifacts;
-    $self->resolve(sub { push @artifacts, $_[0] });
-
-    $self->dump_bootstrap(\@artifacts);
-    $self->save_snapshot(\@artifacts);
-}
-
 sub save_snapshot {
     my($self, $artifacts) = @_;
 
@@ -235,7 +269,8 @@ sub save_snapshot {
             pathname => $artifact->install->{pathname},
             provides => $artifact->provides,
             version => $artifact->version,
-            requirements => $artifact->requirements,
+            # compatibility with Carton snapshot
+            requirements => $artifact->requirements_for([qw( configure build runtime )], ['requires']), 
         );
         $snapshot->add_distribution($dist);
     }
@@ -261,7 +296,7 @@ sub dump_bootstrap {
     }
 
     my $cpanfile = $self->try_cpanfile
-      or Carp::croak "Can't locate 'cpanfile' to load module list.";
+      or die "Can't locate 'cpanfile' to load module list.\n";
 
     my $prereqs = Module::CPANfile->load($cpanfile)->prereqs->as_string_hash;
     my $package = "Carmel::MySetup"; # hide from PAUSE
@@ -366,16 +401,24 @@ sub cmd_tree {
 sub cmd_rollout {
     my $self = shift;
 
-    require ExtUtils::Install;
-    require ExtUtils::InstallPaths;
-
     my @artifacts;
     $self->resolve(sub { push @artifacts, $_[0] });
 
     my $install_base = Path::Tiny->new("local")->absolute;
     $install_base->remove_tree({ safe => 0 }) if $install_base->exists;
 
-    for my $artifact (@artifacts) {
+    $self->rollout_to($install_base, \@artifacts);
+
+    Path::Tiny->new("local/.carmel")->touch;
+}
+
+sub rollout_to {
+    my($self, $install_base, $artifacts) = @_;
+
+    require ExtUtils::Install;
+    require ExtUtils::InstallPaths;
+
+    for my $artifact (@$artifacts) {
         my $dir = pushd $artifact->path;
 
         my $paths = ExtUtils::InstallPaths->new(install_base => $install_base);
@@ -399,8 +442,6 @@ sub cmd_rollout {
 
         select $old unless $self->verbose;
     }
-
-    Path::Tiny->new("local/.carmel")->touch;
 }
 
 sub cmd_package {
@@ -464,7 +505,7 @@ sub build_index {
     $self->resolve(sub {
         my $artifact = shift;
         while (my($pkg, $data) = each %{$artifact->provides}) {
-            my $package = Carton::Package->new($pkg, $data->{version} || 'undef', $artifact->install->{pathname});
+            my $package = Carton::Package->new($pkg, $data->{version}, $artifact->install->{pathname});
             $index->add_package($package);
         }
     });
@@ -528,7 +569,7 @@ sub build_requirements {
     my $self = shift;
 
     my $cpanfile = $self->try_cpanfile
-      or Carp::croak "Can't locate 'cpanfile' to load module list.";
+      or die "Can't locate 'cpanfile' to load module list.\n";
 
     return Module::CPANfile->load($cpanfile)
       ->prereqs->merged_requirements(['runtime', 'test', 'develop'], ['requires']);
@@ -550,7 +591,7 @@ sub merge_requirements {
 }
 
 sub resolve_recursive {
-    my($self, $root_reqs, $requirements, $seen, $cb, $missing_cb, $depth) = @_;
+    my($self, $root_reqs, $requirements, $snapshot, $seen, $cb, $missing_cb, $strict, $depth) = @_;
 
     # TODO rather than mutating $root_reqs directly, we should create a new object
     # that allows accessing the result $requirements
@@ -558,12 +599,20 @@ sub resolve_recursive {
         next if $module eq 'perl';
 
         my $want_version = $root_reqs->requirements_for_module($module);
-        next if $self->is_core($module, $want_version);
 
-        warn "$depth: Resolving $module ($want_version)\n" if $self->verbose;
+        my $artifact;
+        my $dist;
+        if ($dist = $self->find_in_snapshot($snapshot, $module, $root_reqs)) {
+            $artifact = $self->repo->find_match($module, sub { $_[0]->distname eq $dist->name });
+            $artifact ||= $self->repo->find($module, $want_version) unless $strict;
+        } elsif ($self->is_core($module, $want_version)) {
+            next;
+        } else {
+            $artifact = $self->repo->find($module, $want_version);
+        }
 
         # FIXME there's a chance different version of the same module can be loaded here
-        if (my $artifact = $self->repo->find($module, $want_version)) {
+        if ($artifact) {
             warn sprintf "   %s (%s) in %s\n", $module, $artifact->version_for($module), $artifact->path if $self->verbose;
             next if $seen->{$artifact->path}++;
             $cb->($artifact, $depth);
@@ -571,21 +620,76 @@ sub resolve_recursive {
             my $reqs = $artifact->requirements;
             $self->merge_requirements($root_reqs, $reqs, $artifact->distname);
 
-            $self->resolve_recursive($root_reqs, $reqs, $seen, $cb, $missing_cb, $depth + 1);
+            $self->resolve_recursive($root_reqs, $reqs, $snapshot, $seen, $cb, $missing_cb, $strict, $depth + 1);
         } else {
-            $missing_cb->($module, $want_version, $depth);
+            $missing_cb->($module, $want_version, $dist, $depth);
         }
     }
 }
 
 sub resolve {
-    my($self, $cb, $missing_cb) = @_;
+    my($self, $cb, $missing_cb, $strict, $no_snapshot) = @_;
     $missing_cb ||= sub {
-        my($module, $want_version, $depth) = @_;
+        my($module, $want_version, $dist, $depth) = @_;
         die "Can't find an artifact for $module => $want_version\n" .
             "You need to run `carmel install` first to get the modules installed and artifacts built.\n";
     };
-    $self->resolve_recursive($self->requirements, $self->requirements->clone, {}, $cb, $missing_cb, 0);
+
+    my $snapshot = $no_snapshot ? undef : $self->snapshot;
+
+    $self->resolve_recursive($self->requirements, $self->requirements->clone, $snapshot,
+                             {}, $cb, $missing_cb, $strict, 0);
+}
+
+sub find_in_snapshot {
+    my($self, $snapshot, $module, $reqs) = @_;
+
+    return unless $snapshot;
+
+    if (my $dist = $snapshot->find($module)) {
+        warn "@{[$dist->name]} found in snapshot for $module\n" if $self->verbose;
+        if ($self->accepts_all($reqs, $dist)) {
+            return $dist;
+        }
+    }
+
+    warn "$module not found in snapshot\n" if $self->verbose;
+
+    return;
+}
+
+sub accepts_all {
+    my($self, $reqs, $dist) = @_;
+
+    my @packages = keys %{$dist->provides};
+
+    for my $pkg (@packages) {
+        my $version = $dist->provides->{$pkg}{version} || '0';
+        return unless $reqs->accepts_module($pkg, $version);
+    }
+
+    return 1;
+}
+
+sub accepts {
+    my($self, $module, $want_version, $version) = @_;
+
+    CPAN::Meta::Requirements->from_string_hash({ $module => $want_version })
+        ->accepts_module($module, $version || '0');
+}
+
+sub snapshot {
+    my $self = shift;
+
+    my $cpanfile = $self->try_cpanfile;
+    if ($cpanfile && -e "$cpanfile.snapshot") {
+        require Carton::Snapshot;
+        my $snapshot = Carton::Snapshot->new(path => "$cpanfile.snapshot");
+        $snapshot->load;
+        return $snapshot;
+    }
+
+    return;
 }
 
 1;
